/* ****************************************************************************
 * Copyright (C) 2024 Thomas Touhey <thomas@touhey.fr>
 *
 * This software is governed by the CeCILL 2.1 license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/or redistribute the software under the terms of the CeCILL 2.1 license
 * as circulated by CEA, CNRS and INRIA at the following
 * URL: https://cecill.info
 *
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean
 * that it is complicated to manipulate, and that also therefore means that it
 * is reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL 2.1 license and that you accept its terms.
 * ************************************************************************* */

#ifndef INTERNALS_H
#define INTERNALS_H 1
#include <cahute.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(_WIN32) || defined(_WIN64) || defined(__WINDOWS__)
# define WIN32_ENABLED 1
#else
# define WIN32_ENABLED 0
#endif

#if WIN32_ENABLED
# define POSIX_ENABLED 0
#elif defined(__unix__) && __unix__ \
    || (defined(__APPLE__) || defined(__MACH__))
# define POSIX_ENABLED 1
#else
# define POSIX_ENABLED 0
#endif

#if POSIX_ENABLED
# include <fcntl.h>
# include <sys/ioctl.h>
# include <sys/stat.h>
# include <termios.h>
# include <unistd.h>
#endif

#if LIBUSB_ENABLED
# include <libusb.h>
#endif

/* ---
 * Logging internals.
 * --- */

CAHUTE_EXTERN(void)
cahute_log_message(
    int cahute__loglevel,
    char const *cahute__func,
    char const *cahute__format,
    ...
);
CAHUTE_EXTERN(void)
cahute_log_memory(
    int cahute__loglevel,
    char const *cahute__func,
    void const *cahute__memory,
    size_t cahute__size
);

#if defined(__cplusplus) \
    ? CAHUTE_GNUC_PREREQ(2, 6) \
    : !defined(__STRICT_ANSI__) && CAHUTE_GNUC_PREREQ(2, 4)
# define CAHUTE_LOGFUNC __PRETTY_FUNCTION__
#elif defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
# define CAHUTE_LOGFUNC __func__
#elif !defined(__STRICT_ANSI__) && CAHUTE_GNUC_PREREQ(2, 0)
# define CAHUTE_LOGFUNC __FUNCTION__
#else
# define CAHUTE_LOGFUNC NULL
#endif

#define ll_info  10, CAHUTE_LOGFUNC
#define ll_warn  20, CAHUTE_LOGFUNC
#define ll_error 30, CAHUTE_LOGFUNC
#define ll_fatal 40, CAHUTE_LOGFUNC

#define msg cahute_log_message
#define mem cahute_log_memory

/* Macro to print a message and return CAHUTE_ERROR_IMPL.
 * This is necessary to avoid having to track down exactly what was not
 * implemented in the chain using the message.
 * Usage of this macro is enforced with pre-commit. */
#define CAHUTE_RETURN_IMPL(MESSAGE) \
    { \
        msg(ll_error, MESSAGE); \
        return CAHUTE_ERROR_IMPL /* Comment to prevent match by hook. */; \
    }

#if WIN32_ENABLED
/**
 * Log a Windows API error.
 *
 * This is implemented as a separate function to the rest, because gathering
 * an error message for a given error code is quite lengthy.
 *
 * @param func_name Name of the function from which the log is emitted.
 * @param win_func Name of the Windows API function that returned the
 *        error.
 * @param code Windows API error code that was actually returned.
 */
CAHUTE_INLINE(void)
cahute__log_win_error(
    char const *func_name,
    char const *win_func,
    DWORD code
) {
    char buf[1024];
    DWORD buf_size;

    buf_size = FormatMessage(
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        code,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        buf,
        1023,
        NULL
    );

    if (!buf_size) {
        cahute_log_message(
            30,
            func_name,
            "Error 0x%08lX occurred in %s.",
            code,
            win_func
        );
        return;
    }

    buf[buf_size] = '\0';
    cahute_log_message(
        30,
        func_name,
        "Error 0x%08lX occurred in %s: %s",
        code,
        win_func,
        buf
    );
}

# define log_windows_error(FUNC, CODE) \
     cahute__log_win_error(CAHUTE_LOGFUNC, FUNC, CODE)
#endif

/* ---
 * Link internals.
 * --- */

#define CAHUTE_LINK_MEDIUM_READ_BUFFER_SIZE 32768

/* Flags that can be present on a link at runtime. */
#define CAHUTE_LINK_FLAG_CLOSE_MEDIUM 0x00000001
#define CAHUTE_LINK_FLAG_TERMINATE    0x00000002 /* Should terminate. */
#define CAHUTE_LINK_FLAG_RECEIVER     0x00000004 /* Act as a receiver. */

#define CAHUTE_LINK_FLAG_GONE          0x00000100 /* Underlying medium gone. */
#define CAHUTE_LINK_FLAG_TERMINATED    0x00000200 /* Was terminated! */
#define CAHUTE_LINK_FLAG_IRRECOVERABLE 0x00000400 /* Cannot recover. */
#define CAHUTE_LINK_FLAG_ALMODE        0x00000800 /* CAS40 AL data received. */

/* Medium types allowed. */
#if POSIX_ENABLED
# define CAHUTE_LINK_MEDIUM_POSIX_SERIAL 1
#endif
#if WIN32_ENABLED
# define CAHUTE_LINK_MEDIUM_WIN32_SERIAL 2
# define CAHUTE_LINK_MEDIUM_WIN32_CESG   3
# define CAHUTE_LINK_MEDIUM_WIN32_UMS    4
#endif
#if LIBUSB_ENABLED
# define CAHUTE_LINK_MEDIUM_LIBUSB     5
# define CAHUTE_LINK_MEDIUM_LIBUSB_UMS 6
#endif

/* Protocol selection for 'initialize_link_protocol()'. */
#define CAHUTE_LINK_PROTOCOL_SERIAL_AUTO      0
#define CAHUTE_LINK_PROTOCOL_SERIAL_CASIOLINK 1
#define CAHUTE_LINK_PROTOCOL_SERIAL_SEVEN     2
#define CAHUTE_LINK_PROTOCOL_SERIAL_SEVEN_OHP 3
#define CAHUTE_LINK_PROTOCOL_USB_SEVEN        4
#define CAHUTE_LINK_PROTOCOL_USB_SEVEN_OHP    5
#define CAHUTE_LINK_PROTOCOL_USB_MASS_STORAGE 6

/* CASIOLINK variant selection for the same function. */
#define CAHUTE_CASIOLINK_VARIANT_AUTO   0
#define CAHUTE_CASIOLINK_VARIANT_CAS40  1
#define CAHUTE_CASIOLINK_VARIANT_CAS50  2
#define CAHUTE_CASIOLINK_VARIANT_CAS100 3

#if defined(CAHUTE_LINK_MEDIUM_POSIX_SERIAL)
/**
 * POSIX medium state.
 *
 * @property fd File descriptor on the opened medium.
 */
struct cahute_link_posix_medium_state {
    int fd;
};
#endif

#if defined(CAHUTE_LINK_MEDIUM_WIN32_SERIAL) \
    || defined(CAHUTE_LINK_MEDIUM_WIN32_CESG)
/**
 * Windows API medium state.
 *
 * @property handle Device handle.
 * @property overlapped Overlapped I/O adapter.
 */
struct cahute_link_windows_medium_state {
    HANDLE handle;
    OVERLAPPED overlapped;
};
#endif

#if defined(CAHUTE_LINK_MEDIUM_WIN32_UMS)
/**
 * Windows API UMS (SCSI) medium state.
 *
 * @property handle Device handle.
 */
struct cahute_link_windows_ums_medium_state {
    HANDLE handle;
};
#endif

#if defined(CAHUTE_LINK_MEDIUM_LIBUSB)
/**
 * libusb device medium state.
 *
 * @property context libusb context to close once the link is closed.
 * @property handle libusb device handle which to use to make USB requests.
 * @property bulk_in Bulk IN endpoint address to use for reading.
 * @property bulk_out Bulk OUT endpoint address to use for writing.
 */
struct cahute_link_libusb_medium_state {
    libusb_context *context;
    libusb_device_handle *handle;
    int bulk_in;
    int bulk_out;
};
#endif

/**
 * Medium state, to be used depending on the link flags regarding the medium.
 *
 * @property posix Medium state if the selected medium type is POSIX_SERIAL.
 * @property windows Medium state if the selected medium type is WIN32_SERIAL
 *           or WIN32_CESG.
 * @property windows_ums Medium state if the selected medium type is
 *           WIN32_UMS (SCSI over a Windows HANDLE).
 * @property libusb Medium state if the selected medium type is LIBUSB.
 */
union cahute_link_medium_state {
#if defined(CAHUTE_LINK_MEDIUM_POSIX_SERIAL)
    struct cahute_link_posix_medium_state posix;
#endif
#if defined(CAHUTE_LINK_MEDIUM_WIN32_SERIAL) \
    || defined(CAHUTE_LINK_MEDIUM_WIN32_CESG)
    struct cahute_link_windows_medium_state windows;
#endif
#if defined(CAHUTE_LINK_MEDIUM_WIN32_UMS)
    struct cahute_link_windows_ums_medium_state windows_ums;
#endif
#if defined(CAHUTE_LINK_MEDIUM_LIBUSB)
    struct cahute_link_libusb_medium_state libusb;
#endif
};

/* Absolute minimum buffer size for CASIOLINK. */
#define CASIOLINK_MINIMUM_BUFFER_SIZE 50

/* Raw device information size for CASIOLINK, most specifically the
 * CAS100 variant. */
#define CASIOLINK_RAW_DEVICE_INFO_BUFFER_SIZE 33

/* Flags to describe whether device information was obtained or not. */
#define CASIOLINK_FLAG_DEVICE_INFO_OBTAINED 0x00000001

/* Maximum size of raw data that can come from an extended packet.
 * Calculators support data packets with up to 256 raw bytes (512 encoded
 * bytes), but fxRemote uses payloads that go up to 1028 raw bytes
 * (2056 encoded bytes). */
#define SEVEN_MAX_PACKET_DATA_SIZE         1028
#define SEVEN_MAX_ENCODED_PACKET_DATA_SIZE 2056 /* Max data size x 2. */
#define SEVEN_MAX_PACKET_SIZE              2066 /* Enc. data size + 10. */

/* Size of the raw device information buffer for Protocol 7.00.
 * This actually varies between devices: the fx-9860G use 164 bytes,
 * the fx-CG use 188 bytes. */
#define SEVEN_RAW_DEVICE_INFO_BUFFER_SIZE 200

/* Flag to describe whether device information has been requested. */
#define SEVEN_FLAG_DEVICE_INFO_REQUESTED 0x00000001

/**
 * CASIOLINK peer state.
 *
 * @property flags Flags for the CASIOLINK peer state.
 * @property variant Variant with which to force data frame interpretation.
 * @property last_variant Variant for the last data frame.
 * @property raw_device_info Raw device information buffer, so that data
 *           can be extracted later if actual device information is requested.
 */
struct cahute_casiolink_state {
    unsigned long flags;

    int variant;
    int last_variant;

    cahute_u8 raw_device_info[CASIOLINK_RAW_DEVICE_INFO_BUFFER_SIZE];
};

/**
 * Protocol 7.00 peer state.
 *
 * @property flags Flags for the Protocol 7.00 peer state.
 * @property last_command Code of the last executed command. Protocol 7.00
 *           requires the code of the corresponding command to be placed as
 *           the subtype of subsequent data packets.
 * @property last_packet_type Type of the last received packet, or -1 if not
 *           available.
 * @property last_packet_subtype Subtype of the last received packet, or -1
 *           if not available.
 * @property last_packet_data Buffer to the last packet data.
 * @property last_packet_data_size Size of the last packet data.
 * @property raw_device_info Raw device information buffer, so that data can
 *           be extracted later if actual device information is requested.
 * @property raw_device_info_size Raw device information size (not capacity).
 */
struct cahute_seven_state {
    unsigned long flags;

    int last_command;

    int last_packet_type;
    int last_packet_subtype;

    size_t last_packet_data_size;
    size_t raw_device_info_size;

    cahute_u8 last_packet_data[SEVEN_MAX_PACKET_DATA_SIZE];
    cahute_u8 raw_device_info[SEVEN_RAW_DEVICE_INFO_BUFFER_SIZE];
};

/**
 * Protocol 7.00 screenstreaming receiver state.
 *
 * If reception of a frame packet has been successful, the data buffer
 * will contain the frame data.
 *
 * @property last_packet_type Type of the last received packet, or -1 if not
 *           available.
 * @property last_packet_subtype Subtype of the last received packet, if
 *           relevant.
 * @property picture_format Type of the last received picture, as a
 *           ``CAHUTE_PICTURE_FORMAT_*`` constant.
 * @property picture_width Width of the last picture in pixels, -1
 *           if not relevant.
 * @property picture_height Height of the last picture in pixels, -1
 *           if not relevant.
 */
struct cahute_seven_ohp_state {
    int last_packet_type;
    int picture_format;
    int picture_width;
    int picture_height;

    cahute_u8 last_packet_subtype[5];
};

/**
 * Link protocol client state, to be used depending on the protocol selected
 * in the link flags.
 *
 * @property casiolink CASIOLINK peer state.
 * @property seven Protocol 7.00 peer state.
 * @property seven_ohp Protocol 7.00 screenstreaming receiver state.
 */
union cahute_link_protocol_state {
    struct cahute_casiolink_state casiolink;
    struct cahute_seven_state seven;
    struct cahute_seven_ohp_state seven_ohp;
};

/**
 * Internal link representation.
 *
 * @property flags Link flags, as OR'd ``CAHUTE_LINK_FLAG_*`` constants.
 * @property medium Medium type, as any ``CAHUTE_LINK_MEDIUM_*`` constant
 *           representing the medium state to use and how to use it.
 * @property protocol Protocol type, as any ``CAHUTE_LINK_PROTOCOL_*`` constant
 *           representing the protocol state to use.
 * @property serial_flags Current serial flags, as or'd
 *           ``CAHUTE_SERIAL_FLAG_*`` constants.
 * @property serial_speed Current serial speed.
 * @property medium_state State of the specific medium to use, e.g. opened
 *           handles and contexts to close at link closing.
 *           The read buffer is not included within this property.
 * @property protocol_state State of the specific protocol to use, e.g.
 *           current role in the protocol and details regarding the last
 *           received packet.
 *           The protocol data buffer is not included within this property.
 * @property data_buffer General-purpose buffer for the protocol
 *           implementation to use. This can contain payloads, frame data,
 *           etc.
 * @property data_buffer_size Size of the data currently present within
 *           the data buffer, in bytes.
 * @property data_buffer_capacity Total amount of data the data buffer
 *           can contain, in bytes.
 * @property medium_read_buffer Buffer for reading from the medium in a
 *           stream-like interface. See ``cahute_read_from_link`` definition
 *           for more information.
 * @property medium_read_start Offset at which the unread data starts within
 *           the read buffer for the medium.
 * @property medium_read_size Number of unread bytes in the read buffer for the
 *           medium, starting at the offset stored in ``medium_read_start``.
 * @property cached_device_info Device information, if it has been requested
 *           at least once, so it can be free'd when the link is closed.
 */
struct cahute_link {
    unsigned long flags;
    int medium, protocol;
    unsigned long serial_flags;
    unsigned long serial_speed;

    union cahute_link_medium_state medium_state;
    union cahute_link_protocol_state protocol_state;

    cahute_device_info *cached_device_info;

    /* Raw data buffer, used by the protocol implementation to store raw data.
     * This can be of varying length depending on the protocol in use.
     * The buffer is allocated in the same block as the link. */
    cahute_u8 *data_buffer;
    size_t data_buffer_size, data_buffer_capacity;

    /* Stored frame, so that screen reception does not use dynamic
     * memory allocation for every frame. */
    cahute_frame stored_frame;

    /* Read buffer. See ``cahute_read`` definition for more information. */
    size_t medium_read_start, medium_read_size;
    cahute_u8 *medium_read_buffer;
};

/* ---
 * Miscellaneous functions, defined in misc.c
 * --- */

CAHUTE_EXTERN(int) cahute_sleep(unsigned long ms);
CAHUTE_EXTERN(int) cahute_monotonic(unsigned long *msp);

/* ---
 * Link medium functions, defined in medium.c
 * --- */

CAHUTE_EXTERN(int)
cahute_read_from_link(
    cahute_link *link,
    cahute_u8 *buf,
    size_t size,
    unsigned long first_timeout,
    unsigned long next_timeout
);

CAHUTE_EXTERN(int)
cahute_skip_from_link(
    cahute_link *link,
    size_t size,
    unsigned long first_timeout,
    unsigned long next_timeout
);

CAHUTE_EXTERN(int)
cahute_write_to_link(cahute_link *link, cahute_u8 const *buf, size_t size);

CAHUTE_EXTERN(int)
cahute_set_serial_params_to_link(
    cahute_link *link,
    unsigned long flags,
    unsigned long speed
);

CAHUTE_EXTERN(int)
cahute_scsi_request_to_link(
    cahute_link *link,
    cahute_u8 const *command,
    size_t command_size,
    cahute_u8 const *data,
    size_t data_size,
    int *statusp
);

CAHUTE_EXTERN(int)
cahute_scsi_request_from_link(
    cahute_link *link,
    cahute_u8 const *command,
    size_t command_size,
    cahute_u8 *buf,
    size_t buf_size,
    int *statusp
);

/* ---
 * CASIOLINK protocol and file format functions, defined in casiolink.c
 * --- */

CAHUTE_EXTERN(int) cahute_casiolink_initiate(cahute_link *link);

CAHUTE_EXTERN(int) cahute_casiolink_terminate(cahute_link *link);

CAHUTE_EXTERN(int)
cahute_casiolink_receive_data(
    cahute_link *link,
    cahute_data **datap,
    unsigned long timeout
);

CAHUTE_EXTERN(int)
cahute_casiolink_receive_screen(
    cahute_link *link,
    cahute_frame *frame,
    unsigned long timeout
);

CAHUTE_EXTERN(int)
cahute_casiolink_make_device_info(
    cahute_link *link,
    cahute_device_info **infop
);

/* ---
 * Protocol 7.00 functions, defined in seven.c
 * --- */

CAHUTE_EXTERN(int) cahute_seven_initiate(cahute_link *link);

CAHUTE_EXTERN(int) cahute_seven_terminate(cahute_link *link);

CAHUTE_EXTERN(int) cahute_seven_discover(cahute_link *link);

CAHUTE_EXTERN(int)
cahute_seven_receive_data(
    cahute_link *link,
    cahute_data **datap,
    unsigned long timeout
);

CAHUTE_EXTERN(int)
cahute_seven_negotiate_serial_params(
    cahute_link *link,
    unsigned long flags,
    unsigned long speed
);

CAHUTE_EXTERN(int)
cahute_seven_make_device_info(cahute_link *link, cahute_device_info **infop);

CAHUTE_EXTERN(int)
cahute_seven_request_storage_capacity(
    cahute_link *link,
    char const *storage,
    unsigned long *capacityp
);

CAHUTE_EXTERN(int)
cahute_seven_send_file_to_storage(
    cahute_link *link,
    unsigned long flags,
    char const *directory,
    char const *name,
    char const *storage,
    FILE *filep,
    size_t file_size,
    cahute_confirm_overwrite_func *overwrite_func,
    void *overwrite_cookie,
    cahute_progress_func *progress_func,
    void *progress_cookie
);

CAHUTE_EXTERN(int)
cahute_seven_request_file_from_storage(
    cahute_link *link,
    char const *directory,
    char const *name,
    char const *storage,
    FILE *filep,
    cahute_progress_func *progress_func,
    void *progress_cookie
);

CAHUTE_EXTERN(int)
cahute_seven_copy_file_on_storage(
    cahute_link *link,
    char const *source_directory,
    char const *source_name,
    char const *target_directory,
    char const *target_name,
    char const *storage
);

CAHUTE_EXTERN(int)
cahute_seven_delete_file_from_storage(
    cahute_link *link,
    char const *directory,
    char const *name,
    char const *storage
);

CAHUTE_EXTERN(int)
cahute_seven_list_storage_entries(
    cahute_link *link,
    char const *storage,
    cahute_list_storage_entry_func *callback,
    void *cookie
);

CAHUTE_EXTERN(int)
cahute_seven_reset_storage(cahute_link *link, char const *storage);

CAHUTE_EXTERN(int)
cahute_seven_optimize_storage(cahute_link *link, char const *storage);

CAHUTE_EXTERN(int)
cahute_seven_backup_rom(
    cahute_link *link,
    cahute_u8 **romp,
    size_t *sizep,
    cahute_progress_func *progress_func,
    void *progress_cookie
);

CAHUTE_EXTERN(int)
cahute_seven_upload_and_run_program(
    cahute_link *link,
    cahute_u8 const *program,
    size_t program_size,
    unsigned long load_address,
    unsigned long start_address,
    cahute_progress_func *progress_func,
    void *progress_cookie
);

CAHUTE_EXTERN(int)
cahute_seven_flash_system_using_fxremote_method(
    cahute_link *link,
    unsigned long flags,
    cahute_u8 const *system,
    size_t system_size
);

/* ---
 * Protocol 7.00 Screenstreaming functions, defined in seven_ohp.c
 * --- */

CAHUTE_EXTERN(int)
cahute_seven_ohp_receive_screen(
    cahute_link *link,
    cahute_frame *frame,
    unsigned long timeout
);

/* ---
 * MCS encoding and decoding functions, defined in mcs.c
 * --- */

CAHUTE_EXTERN(int)
cahute_mcs_decode_data(
    cahute_data **datap,
    cahute_u8 const *group,
    size_t group_size,
    cahute_u8 const *directory,
    size_t directory_size,
    cahute_u8 const *name,
    size_t name_size,
    cahute_u8 const *content,
    size_t content_size,
    int data_type
);

#endif /* INTERNALS_H */
